// @HEADER
// ************************************************************************
//
//        Phalanx: A Partial Differential Equation Field Evaluation 
//       Kernel for Flexible Management of Complex Dependency Chains
//                    Copyright 2008 Sandia Corporation
//
// Under terms of Contract DE-AC04-94AL85000, there is a non-exclusive
// license for use of this work by or on behalf of the U.S. Government.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are
// met:
//
// 1. Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
//
// 2. Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
//
// 3. Neither the name of the Corporation nor the names of the
// contributors may be used to endorse or promote products derived from
// this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY SANDIA CORPORATION "AS IS" AND ANY
// EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL SANDIA CORPORATION OR THE
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// Questions? Contact Roger Pawlowski (rppawlo@sandia.gov), Sandia
// National Laboratories.
//
// ************************************************************************
// @HEADER


#ifndef MOCK_OBJECT_HPP
#define MOCK_OBJECT_HPP

#include <iostream>

template<typename Scalar> class MockObject;

/** \brief Class to mimic a sacado::DFad object

    The DFad object allocates its own internal memory so its dtor need
    to be called to properly delete all data.  This object is used to
    prove that the dtors are called correctly on each chunk of the
    array.

    NOTE: If you really want the benefits of a contiguous block of
    memory in cache by using this allocator, you should never use a
    DFad object since the internal allocation will bring in
    non-contiguous memory.  You should be using something like SFad.
*/
template<typename Scalar>
class MockObject {

private:

  Scalar m_value;
  Scalar* m_derivative_values;
  int m_derivative_size;

public:
  
  static int m_num_times_empty_ctor_called;
  static int m_num_times_sized_ctor_called;
  static int m_num_times_copy_ctor_called;
  static int m_num_times_assignment_op_called;
  static int m_num_times_dtor_called;

public:

  MockObject() : 
    m_derivative_values(NULL), 
    m_derivative_size(0)
  {++m_num_times_empty_ctor_called;}
  
  MockObject(int derivative_size) : 
    m_derivative_values(NULL),
    m_derivative_size(derivative_size) 
  {
    m_derivative_values = new Scalar[m_derivative_size];
    ++m_num_times_sized_ctor_called;
  }

  MockObject(const MockObject& other) 
  {
    m_derivative_values = NULL;
    m_derivative_size = other.m_derivative_size;
    if (m_derivative_size > 0)
      m_derivative_values = new Scalar[m_derivative_size];
    ++m_num_times_copy_ctor_called;
  }
  
  MockObject& operator=(const MockObject& other)
  {
    if (m_derivative_values != NULL) {
      delete[] m_derivative_values;
      m_derivative_values = NULL;
    }
    m_derivative_size = other.m_derivative_size;
    if (m_derivative_size > 0)
      m_derivative_values = new Scalar[m_derivative_size];
    ++m_num_times_assignment_op_called;
    return *this;
  }

  ~MockObject()
  {
    delete[] m_derivative_values;
    ++m_num_times_dtor_called;
  }

private:


};

template<typename Scalar>
void print_num_ctor_dtor_calls(std::ostream& os)
{
  os << "    # times empty ctor called = " 
     << MockObject<Scalar>::m_num_times_empty_ctor_called << std::endl;
  os << "    # times sized ctor called = " 
     << MockObject<Scalar>::m_num_times_sized_ctor_called << std::endl;
  os << "    # times copy ctor called = " 
     << MockObject<Scalar>::m_num_times_copy_ctor_called << std::endl;
  os << "    # times assignment op called = " 
     << MockObject<Scalar>::m_num_times_assignment_op_called << std::endl;
  os << "    # times dtor called = " 
     << MockObject<Scalar>::m_num_times_dtor_called << std::endl;
}

template<>
int MockObject<int>::m_num_times_empty_ctor_called = 0;
template<>
int MockObject<int>::m_num_times_sized_ctor_called = 0;
template<>
int MockObject<int>::m_num_times_copy_ctor_called = 0;
template<>
int MockObject<int>::m_num_times_assignment_op_called = 0;
template<>
int MockObject<int>::m_num_times_dtor_called = 0;



#endif
