// @HEADER
// ***********************************************************************
//
//          Tpetra: Templated Linear Algebra Services Package
//                 Copyright (2008) Sandia Corporation
//
// Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
// the U.S. Government retains certain rights in this software.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are
// met:
//
// 1. Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
//
// 2. Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
//
// 3. Neither the name of the Corporation nor the names of the
// contributors may be used to endorse or promote products derived from
// this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY SANDIA CORPORATION "AS IS" AND ANY
// EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL SANDIA CORPORATION OR THE
// CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
// EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
// PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
// PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
// NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// Questions? Contact Michael A. Heroux (maherou@sandia.gov)
//
// ************************************************************************
// @HEADER

#ifndef TPETRA_BLOCKMAP_DECL_HPP
#define TPETRA_BLOCKMAP_DECL_HPP

/// \file Tpetra_BlockMap_decl.hpp
/// \brief Declarations for the class Tpetra::BlockMap.

#include <Tpetra_ConfigDefs.hpp>

#ifndef HAVE_TPETRA_CLASSIC_VBR
#  error "It is an error to include this file if VBR (variable-block-size) sparse matrix support is disabled in Tpetra.  If you would like to enable VBR support, please reconfigure Trilinos with the CMake option Tpetra_ENABLE_CLASSIC_VBR set to ON, and rebuild Trilinos."
#else

#include <map>
#include <Tpetra_Map.hpp>

namespace Tpetra {

/** \brief Block-entry counterpart to Tpetra::Map.

  BlockMap doesn't inherit Tpetra::Map, but always holds a Tpetra::Map
  as a class-member attribute.

  Tpetra::BlockMap holds information about how the point entries in
  Tpetra::Map are grouped together in blocks. A block entry consists
  of 1 or more point entries.

  Example usage: If a solution space consists of multiple
  degrees-of-freedom at each finite-element node in a mesh, such as a
  displacement vector, it might be described as having a block of size
  3 (in 3D) at each mesh node.  Thus for a mesh with N nodes, the
  point-entry map will have N*3 entries, whereas the block-map will
  have N blocks, each of size 3.

  \warning This class is DEPRECATED.  There are known outstanding bugs
    with the current implementations of variable-block-size sparse
    matrices and related classes in Tpetra.
*/
template <class LocalOrdinal = Details::DefaultTypes::local_ordinal_type,
          class GlobalOrdinal = Details::DefaultTypes::global_ordinal_type,
          class Node = Details::DefaultTypes::node_type>
class TPETRA_DEPRECATED BlockMap : public Teuchos::Describable {
public:
  typedef LocalOrdinal  local_ordinal_type;
  typedef GlobalOrdinal global_ordinal_type;
  typedef Node          node_type;
  typedef Map<LocalOrdinal, GlobalOrdinal, Node> point_map_type;

  //! @name Constructor/Destructor Methods
  //@{

  /*! \brief BlockMap constructor specifying numGlobalBlocks and constant blockSize.
   */
  BlockMap (global_size_t numGlobalBlocks,
            LocalOrdinal blockSize,
            GlobalOrdinal indexBase,
            const Teuchos::RCP<const Teuchos::Comm<int> > &comm,
            const Teuchos::RCP<Node> &node = defaultArgNode<node_type> ());

  //! BlockMap constructor specifying num global and local blocks, and constant blockSize.
  BlockMap (global_size_t numGlobalBlocks,
            size_t numLocalBlocks,
            LocalOrdinal blockSize,
            GlobalOrdinal indexBase,
            const Teuchos::RCP<const Teuchos::Comm<int> > &comm,
            const Teuchos::RCP<Node> &node = defaultArgNode<node_type> ());

  /// \brief BlockMap constructor specifying numGlobalBlocks and lists
  ///   of local blocks first-global-point-in-blocks, and blockSizes.
  BlockMap (const global_size_t numGlobalBlocks,
            const Teuchos::ArrayView<const GlobalOrdinal>& myGlobalBlockIDs,
            const Teuchos::ArrayView<const GlobalOrdinal>& myFirstGlobalPointInBlocks,
            const Teuchos::ArrayView<const LocalOrdinal>& myBlockSizes,
            const GlobalOrdinal indexBase,
            const Teuchos::RCP<const Teuchos::Comm<int> > &comm,
            const Teuchos::RCP<Node> &node = defaultArgNode<node_type> ());

  /// \brief BlockMap constructor which takes a point-entry Map.
  ///
  /// The arrays myGlobalBlockIDs and myBlockSizes must be the same
  /// length, and sum(myBlockSizes) must equal
  /// pointMap->getNodeNumElements().
  BlockMap (const Teuchos::RCP<const point_map_type>& pointMap,
            const Teuchos::ArrayView<const GlobalOrdinal>& myGlobalBlockIDs,
            const Teuchos::ArrayView<const LocalOrdinal>& myBlockSizes);

  /// \brief BlockMap constructor which takes a point-entry Map.
  ///
  /// The arrays myGlobalBlockIDs and myBlockSizes must be the same
  /// length, and sum(myBlockSizes) must equal
  /// pointMap->getNodeNumElements().
  ///
  /// \warning This constructor is DEPRECATED, because there is never
  ///   a reason to use a Node instance other than the one provided by
  ///   the point Map.
  TPETRA_DEPRECATED
  BlockMap (const Teuchos::RCP<const point_map_type>& pointMap,
            const Teuchos::ArrayView<const GlobalOrdinal>& myGlobalBlockIDs,
            const Teuchos::ArrayView<const LocalOrdinal>& myBlockSizes,
            const Teuchos::RCP<Node>& node);

  //! Destructor.
  ~BlockMap () {}

  //@}
  //! @name Attribute Accessor Methods
  //@{

  //! Return this block Map's point-entry Map.
  Teuchos::RCP<const point_map_type> getPointMap () const {
    return pointMap_;
  }

  //! Return global number of blocks.
  global_size_t getGlobalNumBlocks() const;

  //! Return number of blocks on the local processor.
  size_t getNodeNumBlocks() const;

  //! Return array-view of block-ids for this local processor.
  Teuchos::ArrayView<const GlobalOrdinal> getNodeBlockIDs() const;

  //! Return true if all blocks have the same size.
  bool isBlockSizeConstant() const;

  //! Return ArrayRCP of first-local-point in local blocks.
  Teuchos::ArrayRCP<const LocalOrdinal> getNodeFirstPointInBlocks() const;

  //! Return device-resident ArrayRCP of first-local-point in local blocks.
  /*! This version of this method is primarily used internally by VbrMatrix
      for passing data to the matrix-vector-product kernel.
  */
  Teuchos::ArrayRCP<const LocalOrdinal> getNodeFirstPointInBlocks_Device() const;

  //! Return the globalBlockID corresponding to the given localBlockID
  /*! If localBlockID is not present on this processor, returns Teuchos::OrdinalTraits<LocalOrdinal>::invalid().
  */
  GlobalOrdinal getGlobalBlockID(LocalOrdinal localBlockID) const;

  //! Return the localBlockID corresponding to the given globalBlockID
  /*! If globalBlockID is not present on this processor, returns Teuchos::OrdinalTraits<LocalOrdinal>::invalid().
  */
  LocalOrdinal getLocalBlockID(GlobalOrdinal globalBlockID) const;

  //! Return the block-size for localBlockID
  /*! If localBlockID is out of range (less than 0 or greater/equal num-local-blocks),
   * then std::runtime_error is thrown.
   */
  LocalOrdinal getLocalBlockSize(LocalOrdinal localBlockID) const;

  //! Return the first local point-index corresponding to localBlockID
  /*! If localBlockID is out of range (less than 0 or greater/equal num-local-blocks),
   * then std::runtime_error is thrown.
   */
  LocalOrdinal getFirstLocalPointInLocalBlock(LocalOrdinal localBlockID) const;

  //! Return the first global point-index corresponding to localBlockID
  /*! If localBlockID is out of range (less than 0 or greater/equal num-local-blocks),
   * then std::runtime_error is thrown.
   */
  GlobalOrdinal getFirstGlobalPointInLocalBlock(LocalOrdinal localBlockID) const;

  //! Return the first-global-point-in-block and block-sizes for a list of block-IDs on remote processors.
  void
  getRemoteBlockInfo (const Teuchos::ArrayView<const GlobalOrdinal>& GBIDs,
                      const Teuchos::ArrayView<GlobalOrdinal>& firstGlobalPointInBlocks,
                      const Teuchos::ArrayView<LocalOrdinal>& blockSizes) const;
  //@}

private:
  void
  initWithPointMap (const Teuchos::RCP<const point_map_type>& pointMap,
                    const Teuchos::ArrayView<const GlobalOrdinal>& myGlobalBlockIDs,
                    const Teuchos::ArrayView<const LocalOrdinal>& blockSizes,
                    const Teuchos::RCP<Node>& node);

  void setup_noncontig_mapping();

  Teuchos::RCP<const point_map_type> pointMap_;
  global_size_t globalNumBlocks_;
  Teuchos::Array<GlobalOrdinal> myGlobalBlockIDs_;
  Teuchos::ArrayRCP<LocalOrdinal> pbuf_firstPointInBlock_;
  Teuchos::ArrayRCP<const LocalOrdinal> view_firstPointInBlock_;
  bool blockIDsAreContiguous_;
  LocalOrdinal constantBlockSize_;
  /// \brief Global-to-local index lookup table.
  ///
  /// TODO: Use Tpetra::Details::HashTable here instead.
  std::map<GlobalOrdinal,LocalOrdinal> map_global_to_local_;
};


template<class LocalOrdinal,class GlobalOrdinal,class Node>
Teuchos::RCP<const Tpetra::Map<LocalOrdinal,GlobalOrdinal,Node> >
convertBlockMapToPointMap (const Tpetra::BlockMap<LocalOrdinal,GlobalOrdinal,Node>& blockMap)
{
  using Teuchos::RCP;
  using Teuchos::rcp;
  typedef Map<LocalOrdinal, GlobalOrdinal, Node> point_map_type;
  typedef global_size_t GST;

  const GST numGlobalElems = Teuchos::OrdinalTraits<GST>::invalid ();
  const GlobalOrdinal indexBase = blockMap.getPointMap ()->getIndexBase ();
  RCP<const Teuchos::Comm<int> > comm = blockMap.getPointMap ()->getComm ();
  RCP<Node> node = blockMap.getPointMap ()->getNode ();

  // Create a point-entry map where each point
  // corresponds to a block in the block map.
  return rcp (new point_map_type (numGlobalElems, blockMap.getNodeBlockIDs (),
                                  indexBase, comm, node));
}

} // namespace Tpetra

#endif // ! HAVE_TPETRA_CLASSIC_VBR
#endif // ! TPETRA_BLOCKMAP_DECL_HPP

